/*
 * HAI.cpp
 * 
 * This is a mexable c++ file containing code for computing hierarchy agreement index (HAI) 
 * values between two properly-formatted input arrays, representing two hierarchical
 * clusterings of a given set of instances.
 * 
 * Both input matrices must be n by D integer arrays, where each column represents a single layer
 * of the tree, and each row describes a single point within the cluster hierarchy.  Each element (i, j)
 * contains the positive node id of the tree node containing element i at depth j, or 
 * 0 if element i's leaf node is located at a depth less than j.
 * 
 * This function can either return a single value (a scalar double representing the HAI score between
 * the two inputs) or 3 values [HAIscore, dist1, dist2], where dist1 and dist2 are the hierarchy distance
 * matrices generated by the first and second input hierarchies, respectively.
 */

#include <matrix.h>
#include <mex.h>
#include <vector>
#include <cmath>
#include <unordered_set>

using namespace std;

//First, a number of definitions for needed helper classes and methods
//
//class for representing hierarchies in a tree structure
class hierarchy_node
{
public:
    //constructor - idin = the cluster id for the node
    hierarchy_node(int idin)
    {
        id = idin;
        _leaf = true;
    }

    //recursively insert a point into the hierarchy
    void addmember(mwSize point, const int* clusts, mwSize depthleft, mwSize n)
    {
        int subclust = *clusts;
        _members.insert(point);
        if (subclust > 0 && depthleft > 0)
        {
            //try to find correct subcluster to insert point into
            vector<hierarchy_node>::iterator it;
            bool found = false;
            for (it = _children.begin(); it != _children.end(); it++)
            {
                if (it->id == subclust)
                {
                    it->addmember(point, clusts + n, depthleft - 1, n);
                    found = true;
                    break;
                }
            }
            //if it doesn't exist, create it
            if (!found)
            {
                _leaf = false;
                _children.push_back(hierarchy_node(subclust));
                _children.back().addmember(point, clusts + n, depthleft - 1, n);
            }
        }
    }

    //compute the fractional sizes of this node and all others below it,
    //using the given total value of n
    void compute_sizes(mwSize n)
    {
        fsize = static_cast<double>(size()) / static_cast<double>(n);
        vector<hierarchy_node>::iterator it;
        for (it = _children.begin(); it < _children.end(); it++)
            it->compute_sizes(n);
    }

    //generate an n by n distance matrix based on distance within this subtree
    double* distmat()
    {
        mwSize n = size();
        double* out = new double[n * n];
        mwSize i, j;
        for (i = 0; i < n; i++)
        {
            for (j = i; j < n; j++)
            {
                out[i * n + j] = getdist(i, j);
                ;
            }
        }
        for (i = 0; i < n; i++)
        {
            for (j = i - 1; j >= 0; j--)
            {
                out[i * n + j] = out[j * n + i];
            }
        }
        return out;
    }

    //helper method for distmat - recursively computes the distance between two points
    double getdist(int i, int j)
    {
        if (_leaf)
            return 0.0f;
        for (vector<hierarchy_node>::iterator it = _children.begin(); it != _children.end(); it++)
        {
            int count = it->count(i);
            count += it->count(j);
            if (count == 1)
            {
                return fsize;
            }
            if (count == 2)
            {
                return it->getdist(i, j);
            }
        }
    }

    //simple accessor for size of node
    mwSize size()
    {
        return _members.size();
    }
    //get the number of occurrences of element i (should be 1 or 0)
    int count(int i)
    {
        return _members.count(i);
    }

    //cluster id of this node
    int id;
    //fractional size of this node, calculated by calling compute_sizes()
    double fsize;

private:
    bool _leaf; //tracks whether this node is currently a leaf node
    vector<hierarchy_node> _children; //stores subnodes
    unordered_set<int> _members; //tracks which elements are members of this node
};

//now the actual mex function which computes HAI
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    //check inputs
    if (nlhs > 1 && nlhs != 3)
        mexErrMsgTxt("Wrong number of outputs - HAI returns a single value, or 3 values.");
    if (nrhs != 2)
        mexErrMsgTxt("Wrong number of inputs - HAI requires exactly 2 matrices as input.");

    if (!mxIsInt32(prhs[0]) || mxIsComplex(prhs[0]) || !mxIsInt32(prhs[1]) || mxIsComplex(prhs[1])
            || mxGetNumberOfDimensions(prhs[0]) != 2 || mxGetNumberOfDimensions(prhs[1]) != 2)
        mexErrMsgTxt("Inputs must be real 2d int32 arrays.");

    const mwSize* s1 = mxGetDimensions(prhs[0]);
    const mwSize* s2 = mxGetDimensions(prhs[1]);

    if (!(s1[0] == s2[0]))
        mexErrMsgTxt("Inputs must have the same number of rows.");

    //access/initialize data
    mwSize n = s1[0];
    mwSize D1 = s1[1];
    mwSize D2 = s2[1];
    const int* hierarchy1 = static_cast<int*>(mxGetData(prhs[0]));
    const int* hierarchy2 = static_cast<int*>(mxGetData(prhs[1]));

    //compute HAI
    //first read hierarchies into actual tree structures
    hierarchy_node root1(0);
    hierarchy_node root2(0);
    for (mwSize i = 0; i < n; i++)
    {
        root1.addmember(i, hierarchy1 + i, D1, n);
        root2.addmember(i, hierarchy2 + i, D2, n);
    }
    root1.compute_sizes(n);
    root2.compute_sizes(n);
    //now compute hierarchy distance matrices and leaf node matrices for each hierarchy
    double* dists1 = root1.distmat();
    double* dists2 = root2.distmat();
    //now iterate through matrices and compute total difference between them
    double conflict = 0.0;
    for (mwSize i = 0; i < n * n; i++)
        conflict += abs(dists1[i] - dists2[i]);

    double out = 1.0 - (conflict / (static_cast<double>(n) * n));

    plhs[0] = mxCreateDoubleScalar(out);
    //also return distance matrices if requested
    if (nlhs == 3)
    {
        plhs[1] = mxCreateDoubleMatrix(n, n, mxREAL);
        double* temp = mxGetPr(plhs[1]);
        for (mwSize i = 0; i < n * n; i++)
            temp[i] = dists1[i];
        plhs[2] = mxCreateDoubleMatrix(n, n, mxREAL);
        temp = mxGetPr(plhs[2]);
        for (mwSize i = 0; i < n * n; i++)
            temp[i] = dists2[i];
    }
    delete[] dists1, dists2;
}
